import asyncio
import logging
from typing import Any, Dict, List, Optional

import aiohttp

_KADASTER_SPARQL_ENDPOINT = (
    "https://api.labs.kadaster.nl/datasets/dst/kkg/services/default/sparql"
)

_VERBLIJFSOBJECTEN_QUERY_TEMPLATE = """
PREFIX sor: <https://data.kkg.kadaster.nl/sor/model/def/>
PREFIX nen3610: <https://data.kkg.kadaster.nl/nen3610/model/def/>
PREFIX geo: <http://www.opengis.net/ont/geosparql#>
SELECT DISTINCT ?identificatie ?verblijfsobjectWKT
WHERE {{
  ?verblijfsobject sor:geregistreerdMet/nen3610:identificatie ?identificatie .
  ?verblijfsobject geo:hasGeometry/geo:asWKT ?verblijfsobjectWKT.
  FILTER (?identificatie IN ( {identificaties} ))
}}
"""

_kadaster_semaphore: Optional[asyncio.Semaphore] = None

# Create a module-level logger
logger = logging.getLogger("monumenten.api.kadaster")


def _get_semaphore(loop: asyncio.AbstractEventLoop) -> asyncio.Semaphore:
    global _kadaster_semaphore
    if _kadaster_semaphore is None:
        _kadaster_semaphore = asyncio.Semaphore(4)
    return _kadaster_semaphore


async def _query_verblijfsobjecten(
    session: aiohttp.ClientSession, identificaties: List[str]
) -> List[Dict[str, Any]]:
    async with _get_semaphore(asyncio.get_running_loop()):
        identificaties_str = ", ".join(
            f'"{identificatie}"' for identificatie in identificaties
        )
        query = _VERBLIJFSOBJECTEN_QUERY_TEMPLATE.format(
            identificaties=identificaties_str
        )
        data = {"query": query, "format": "json"}
        retries = 3
        for poging in range(retries):
            try:
                async with session.post(
                    _KADASTER_SPARQL_ENDPOINT, data=data
                ) as response:
                    response.raise_for_status()
                    resultaat = await response.json()
                    if isinstance(resultaat, list):
                        return resultaat
                    else:
                        logger.warning(
                            "Onverwacht response formaat bij poging %d: %s",
                            poging + 1,
                            resultaat,
                        )
            except aiohttp.ClientResponseError as e:
                if poging != retries - 1:
                    logger.warning(
                        "Poging %d/%d voor verblijfsobjecten query mislukt: %s. Opnieuw proberen over 1 seconde...",
                        poging + 1,
                        retries,
                        str(e),
                    )
                    await asyncio.sleep(1)
                else:
                    logger.error(
                        "Alle pogingen voor verblijfsobjecten query mislukt: %s", str(e)
                    )
                    raise
        return List[Dict[str, Any]]()



"""Client voor monumenten package."""

from typing import Any, Dict, List, Optional, cast

import aiohttp
import pandas as pd

from monumenten._processing import _query


class MonumentenClient:
    """Client voor het ophalen van monumentgegevens van verschillende Nederlandse overheids-API's.

    Args:
        session (Optional[aiohttp.ClientSession]): Optionele aiohttp.ClientSession. Indien niet opgegeven wordt
                een nieuwe sessie aangemaakt en beheerd door de client.
    """

    def __init__(self, session: Optional[aiohttp.ClientSession] = None) -> None:
        self._session = session
        self._owns_session = session is None

    async def __aenter__(self) -> "MonumentenClient":
        if self._owns_session:
            self._session = aiohttp.ClientSession()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[Any],
    ) -> None:
        if self._owns_session and self._session:
            await self._session.close()

    async def process_from_df(
        self,
        df: pd.DataFrame,
        verblijfsobject_id_col: str,
    ) -> pd.DataFrame:
        """Verwerk een DataFrame met verblijfsobject ID's.

        Args:
            df (pd.DataFrame): Input DataFrame met verblijfsobject ID's
            verblijfsobject_id_col (str): Naam van de kolom met de verblijfsobject ID's

        Returns:
            pd.DataFrame: DataFrame met toegevoegde monumentinformatie

        Raises:
            RuntimeError: Als de client niet als context manager wordt gebruikt
        """
        if not self._session:
            raise RuntimeError("Client must be used as a context manager")

        results = await _query(
            self._session, df[verblijfsobject_id_col].drop_duplicates()
        )
        merged = pd.merge(
            df,
            results,
            left_on=verblijfsobject_id_col,
            right_on="identificatie",
            how="left",
        )

        if "identificatie" not in df.columns:
            merged = merged.drop(columns=["identificatie"])

        rijksmonument_nummer_position = merged.columns.get_loc("rijksmonument_nummer")
        merged.insert(
            rijksmonument_nummer_position + 1,
            "rijksmonument_url",
            "https://monumentenregister.cultureelerfgoed.nl/monumenten/"
            + merged["rijksmonument_nummer"]
            .fillna("")
            .astype(str)
            .where(merged["rijksmonument_nummer"].notna(), None),
        )

        merged.insert(
            rijksmonument_nummer_position,
            "is_rijksmonument",
            merged["rijksmonument_nummer"].notna(),
        )

        beschermd_gezicht_naam_position = merged.columns.get_loc(
            "beschermd_gezicht_naam"
        )

        merged.insert(
            beschermd_gezicht_naam_position,
            "is_beschermd_gezicht",
            merged["beschermd_gezicht_naam"].notna(),
        )

        return merged

    async def process_from_list(
        self, verblijfsobject_ids: List[str]
    ) -> Dict[str, Dict[str, Any]]:
        """Verwerk een lijst met verblijfsobject ID's.

        Args:
            verblijfsobject_ids (List[str]): Lijst met te verwerken ID's

        Returns:
            Dict[str, Dict[str, Any]]: Dictionary met verblijfsobject ID's als keys en dictionaries met monumentinformatie als values
        """
        df = pd.DataFrame({"bag_verblijfsobject_id": verblijfsobject_ids})
        result = await self.process_from_df(df, "bag_verblijfsobject_id")
        result = result.replace({pd.NA: None, pd.NaT: None})
        return cast(
            Dict[str, Dict[str, Any]],
            result.set_index("bag_verblijfsobject_id").to_dict(orient="index"),
        )



import pandas as pd
import pytest
import pytest_asyncio
from monumenten import MonumentenClient


@pytest_asyncio.fixture(scope="function")
async def client():
    async with MonumentenClient() as client:
        yield client


@pytest.mark.asyncio
async def test_process_from_list(client):
    bag_verblijfsobject_ids = [
        "0599010000360091",
        "0599010000486642",
        "0599010000281115",
    ]

    result = await client.process_from_list(bag_verblijfsobject_ids)

    assert len(result) == 3

    # Test rijksmonument
    assert "0599010000360091" in result
    assert result["0599010000360091"]["is_rijksmonument"] is True
    assert result["0599010000360091"]["rijksmonument_nummer"] == "524327"
    assert (
        result["0599010000360091"]["rijksmonument_url"]
        == "https://monumentenregister.cultureelerfgoed.nl/monumenten/524327"
    )
    assert result["0599010000360091"]["is_beschermd_gezicht"] is False

    # Test non-monument
    assert "0599010000486642" in result
    assert result["0599010000486642"]["is_rijksmonument"] is False
    assert result["0599010000486642"]["rijksmonument_nummer"] is None
    assert result["0599010000486642"]["is_beschermd_gezicht"] is False

    # Test beschermd gezicht
    assert "0599010000281115" in result
    assert result["0599010000281115"]["is_rijksmonument"] is False
    assert result["0599010000281115"]["is_beschermd_gezicht"] is True
    assert result["0599010000281115"]["beschermd_gezicht_naam"] == "Kralingen - Midden"


@pytest.mark.asyncio
async def test_process_from_df(client):
    input_df = pd.DataFrame(
        {
            "bag_verblijfsobject_id": [
                "0599010000360091",  # rijksmonument
                "0599010000486642",  # non-monument
                "0599010000183527",  # both rijksmonument and beschermd gezicht
                "0599010000281115",  # beschermd gezicht only
            ]
        }
    )

    result = await client.process_from_df(
        df=input_df, verblijfsobject_id_col="bag_verblijfsobject_id"
    )

    assert isinstance(result, pd.DataFrame)
    assert len(result) == 4

    # Test rijksmonument
    assert result.iloc[0]["bag_verblijfsobject_id"] == "0599010000360091"
    assert bool(result.iloc[0]["is_rijksmonument"]) is True
    assert result.iloc[0]["rijksmonument_nummer"] == "524327"
    assert (
        result.iloc[0]["rijksmonument_url"]
        == "https://monumentenregister.cultureelerfgoed.nl/monumenten/524327"
    )
    assert bool(result.iloc[0]["is_beschermd_gezicht"]) is False

    # Test non-monument
    assert result.iloc[1]["bag_verblijfsobject_id"] == "0599010000486642"
    assert bool(result.iloc[1]["is_rijksmonument"]) is False
    assert pd.isna(result.iloc[1]["rijksmonument_nummer"])
    assert bool(result.iloc[1]["is_beschermd_gezicht"]) is False

    # Test combined rijksmonument and beschermd gezicht
    assert result.iloc[2]["bag_verblijfsobject_id"] == "0599010000183527"
    assert bool(result.iloc[2]["is_rijksmonument"]) is True
    assert result.iloc[2]["rijksmonument_nummer"] == "32807"
    assert (
        result.iloc[2]["rijksmonument_url"]
        == "https://monumentenregister.cultureelerfgoed.nl/monumenten/32807"
    )
    assert bool(result.iloc[2]["is_beschermd_gezicht"]) is True
    assert result.iloc[2]["beschermd_gezicht_naam"] == "Rotterdam - Scheepvaartkwartier"

    # Test beschermd gezicht only
    assert result.iloc[3]["bag_verblijfsobject_id"] == "0599010000281115"
    assert bool(result.iloc[3]["is_rijksmonument"]) is False
    assert bool(result.iloc[3]["is_beschermd_gezicht"]) is True
    assert result.iloc[3]["beschermd_gezicht_naam"] == "Kralingen - Midden"



"""Interne processing functies voor de monumenten package."""

import asyncio
from typing import Any, Dict, List, Tuple

import aiohttp
import geopandas as gpd
import pandas as pd
from tqdm.asyncio import tqdm_asyncio

from monumenten._api._cultureel_erfgoed import (
    _query_beschermde_gebieden,
    _query_rijksmonumenten,
)
from monumenten._api._kadaster import _query_verblijfsobjecten

_QUERY_BATCH_GROOTTE = 500  # lijkt meest optimaal qua performance


async def _process_batch(
    session: aiohttp.ClientSession, batch: List[str], bg_df: gpd.GeoDataFrame
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]], int]:
    """Verwerk een batch verblijfsobjecten.

    Args:
        session (aiohttp.ClientSession): De sessie voor HTTP requests
        batch (List[str]): Lijst met verblijfsobject ID's
        bg_df (gpd.GeoDataFrame): GeoDataFrame met beschermde gezichten

    Returns:
        Tuple[List[Dict[str, Any]], List[Dict[str, Any]], int]: Tuple met rijksmonumenten,
            beschermde gezichten en aantal verwerkte objecten
    """
    # Get the current event loop
    loop = asyncio.get_running_loop()

    # Create tasks using the current loop
    rijksmonumenten_taak = loop.create_task(_query_rijksmonumenten(session, batch))
    verblijfsobjecten_taak = loop.create_task(_query_verblijfsobjecten(session, batch))

    # Wait for both tasks to complete
    rijksmonumenten, verblijfsobjecten = await asyncio.gather(
        rijksmonumenten_taak, verblijfsobjecten_taak
    )

    # Process 'verblijfsobjecten' into GeoDataFrame
    if verblijfsobjecten:
        vo_df = pd.DataFrame(verblijfsobjecten)
        vo_df["geometry"] = gpd.GeoSeries.from_wkt(vo_df["verblijfsobjectWKT"])
        vo_df = gpd.GeoDataFrame(
            vo_df[["identificatie", "geometry"]], geometry="geometry"
        )
    else:
        vo_df = gpd.GeoDataFrame(columns=["identificatie", "geometry"])

    # Perform spatial join
    if not vo_df.empty:
        joined_df = gpd.sjoin(vo_df, bg_df, how="left", predicate="within")

        verblijfsobjecten_in_beschermd_gezicht = joined_df[
            ["identificatie", "beschermd_gezicht_naam"]
        ].to_dict("records")
    else:
        verblijfsobjecten_in_beschermd_gezicht = list[Dict[str, Any]]()

    return rijksmonumenten, verblijfsobjecten_in_beschermd_gezicht, len(batch)


async def _query(
    session: aiohttp.ClientSession, verblijfsobject_ids: List[str]
) -> pd.DataFrame:
    """Voer queries uit voor een lijst verblijfsobjecten.

    Args:
        session (aiohttp.ClientSession): De sessie voor HTTP requests
        verblijfsobject_ids (List[str]): Lijst met verblijfsobject ID's

    Returns:
        pd.DataFrame: DataFrame met monumentinformatie
    """
    # Load 'beschermde_gebieden' and convert to GeoDataFrame
    beschermde_gebieden = await _query_beschermde_gebieden(session)
    bg_df = gpd.GeoDataFrame()

    if beschermde_gebieden:
        bg_df = pd.DataFrame(beschermde_gebieden)
        bg_df["geometry"] = gpd.GeoSeries.from_wkt(bg_df["gezichtWKT"])
        bg_df = gpd.GeoDataFrame(
            bg_df[["beschermd_gezicht_naam", "geometry"]], geometry="geometry"
        )

    # Ensure spatial index is built
    bg_df.sindex

    rijksmonumenten_result = list[Dict[str, Any]]()
    verblijfsobjecten_in_beschermd_gezicht_result = list[Dict[str, Any]]()

    # Prepare batches
    batches = [
        verblijfsobject_ids[i : i + _QUERY_BATCH_GROOTTE]
        for i in range(0, len(verblijfsobject_ids), _QUERY_BATCH_GROOTTE)
    ]

    # Create tasks for each batch
    tasks = [_process_batch(session, batch, bg_df) for batch in batches]

    progress_bar = tqdm_asyncio(total=len(verblijfsobject_ids))

    for task in asyncio.as_completed(tasks):
        rijksmonumenten, verblijfsobjecten, aantal = await task
        rijksmonumenten_result.extend(rijksmonumenten)
        verblijfsobjecten_in_beschermd_gezicht_result.extend(verblijfsobjecten)
        progress_bar.update(aantal)

    progress_bar.close()

    df_rijksmonumenten = (
        pd.DataFrame(rijksmonumenten_result)
        if rijksmonumenten_result
        else pd.DataFrame({"identificatie": [], "rijksmonument_nummer": []})
    )

    df_beschermd_gezicht = (
        pd.DataFrame(verblijfsobjecten_in_beschermd_gezicht_result)
        if verblijfsobjecten_in_beschermd_gezicht_result
        else pd.DataFrame({"identificatie": [], "beschermd_gezicht_naam": []})
    )

    result = pd.merge(
        df_rijksmonumenten, df_beschermd_gezicht, on="identificatie", how="outer"
    )

    return result



import asyncio
import logging
from typing import Any, Dict, List, Optional

import aiohttp

# Create a module-level logger
logger = logging.getLogger("monumenten.api.cultureel_erfgoed")

_CULTUREEL_ERFGOED_SPARQL_ENDPOINT = (
    "https://api.linkeddata.cultureelerfgoed.nl/datasets/rce/cho/sparql"
)

_RIJKSMONUMENTEN_QUERY_TEMPLATE = """
PREFIX ceo:<https://linkeddata.cultureelerfgoed.nl/def/ceo#>
PREFIX bag:<http://bag.basisregistraties.overheid.nl/bag/id/>
PREFIX rn:<https://data.cultureelerfgoed.nl/term/id/rn/>
SELECT ?identificatie (MAX(?nummer) as ?rijksmonument_nummer)
WHERE {{
    ?monument ceo:heeftJuridischeStatus rn:b2d9a59a-fe1e-4552-9a05-3c2acddff864 ;
              ceo:rijksmonumentnummer ?nummer ;
              ceo:heeftBasisregistratieRelatie ?basisregistratieRelatie .
    ?basisregistratieRelatie ceo:heeftBAGRelatie ?bagRelatie .
    ?bagRelatie ceo:verblijfsobjectIdentificatie ?identificatie .
    VALUES ?identificatie {{ {identificaties} }}
}}
GROUP BY ?identificatie
"""

_BESCHERMDE_GEZICHTEN_QUERY = """
PREFIX ceo:<https://linkeddata.cultureelerfgoed.nl/def/ceo#>
PREFIX rn:<https://data.cultureelerfgoed.nl/term/id/rn/>
PREFIX geo: <http://www.opengis.net/ont/geosparql#>
SELECT DISTINCT ?gezicht ?beschermd_gezicht_naam ?gezichtWKT
WHERE {{
  ?gezicht
      ceo:heeftGeometrie ?gezichtGeometrie ;
      ceo:heeftGezichtsstatus rn:fd968529-bf70-4afa-8564-7c6c2fcfcc54;
      ceo:heeftNaam/ceo:naam ?beschermd_gezicht_naam.
  ?gezichtGeometrie geo:asWKT ?gezichtWKT.
}}
"""

_cultureel_erfgoed_semaphore: Optional[asyncio.Semaphore] = None


def _get_semaphore(loop: asyncio.AbstractEventLoop) -> asyncio.Semaphore:
    """
    Verkrijgt een semaphore voor het beperken van gelijktijdige aanvragen naar de Cultureel Erfgoed API.

    Args:
        loop (asyncio.AbstractEventLoop): De asyncio event loop

    Returns:
        asyncio.Semaphore: Een asyncio.Semaphore object dat het aantal gelijktijdige aanvragen beperkt tot 4
    """
    global _cultureel_erfgoed_semaphore
    if _cultureel_erfgoed_semaphore is None:
        _cultureel_erfgoed_semaphore = asyncio.Semaphore(4)
    return _cultureel_erfgoed_semaphore


async def _query_rijksmonumenten(
    session: aiohttp.ClientSession, identificaties: List[str]
) -> List[Dict[str, Any]]:
    """
    Voert een SPARQL-query uit om rijksmonumenten op te halen voor gegeven BAG-identificaties.

    Args:
        session (aiohttp.ClientSession): De aiohttp ClientSession voor het uitvoeren van de HTTP-aanvraag
        identificaties (List[str]): Lijst van BAG-identificaties waarvoor rijksmonumenten worden opgezocht

    Returns:
        List[Dict[str, Any]]: Lijst van dictionaries met informatie over gevonden rijksmonumenten

    Raises:
        aiohttp.ClientResponseError: Bij fouten in de HTTP-aanvraag na 3 pogingen
    """
    async with _get_semaphore(asyncio.get_running_loop()):
        identificaties_str = " ".join(
            f'"{identificatie}"' for identificatie in identificaties
        )
        query = _RIJKSMONUMENTEN_QUERY_TEMPLATE.format(
            identificaties=identificaties_str
        )
        data = {"query": query, "format": "json"}
        retries = 3
        for poging in range(retries):
            try:
                async with session.post(
                    _CULTUREEL_ERFGOED_SPARQL_ENDPOINT, data=data
                ) as response:
                    response.raise_for_status()
                    resultaat = await response.json()
                    if isinstance(resultaat, list):
                        return resultaat
                    else:
                        logger.warning(
                            "Unexpected response format on attempt %d: %s",
                            poging + 1,
                            resultaat,
                        )
            except aiohttp.ClientResponseError as e:
                if poging != retries - 1:
                    logger.warning(
                        "Poging %d/%d voor rijksmonumenten query mislukt: %s. Opnieuw proberen over 1 seconde...",
                        poging + 1,
                        retries,
                        str(e),
                    )
                    await asyncio.sleep(1)
                else:
                    raise
        return List[Dict[str, Any]]()


async def _query_beschermde_gebieden(
    session: aiohttp.ClientSession,
) -> List[Dict[str, Any]]:
    """
    Voert een SPARQL-query uit om beschermde stads- en dorpsgezichten op te halen.

    Args:
        session (aiohttp.ClientSession): De aiohttp ClientSession voor het uitvoeren van de HTTP-aanvraag

    Returns:
        List[Dict[str, Any]]: Lijst van dictionaries met informatie over beschermde stads- en dorpsgezichten

    Raises:
        aiohttp.ClientResponseError: Bij fouten in de HTTP-aanvraag na 3 pogingen
    """
    retries = 3
    for poging in range(retries):
        try:
            data = {"query": _BESCHERMDE_GEZICHTEN_QUERY, "format": "json"}

            async with session.post(
                _CULTUREEL_ERFGOED_SPARQL_ENDPOINT, data=data
            ) as response:
                response.raise_for_status()
                resultaat = await response.json()
                if isinstance(resultaat, list):
                    return resultaat
                else:
                    logger.warning(
                        "Onverwacht response formaat bij poging %d: %s",
                        poging + 1,
                        resultaat,
                    )
        except aiohttp.ClientResponseError as e:
            if poging != retries - 1:
                logger.warning(
                    "Poging %d/%d voor beschermde gebieden query mislukt: %s. Opnieuw proberen over 1 seconde...",
                    poging + 1,
                    retries,
                    str(e),
                )
                await asyncio.sleep(1)
            else:
                raise
    return List[Dict[str, Any]]()
